#pragma once

#include <dune/common/parallel/mpihelper.hh>

#include <dune/grid/common/gridenums.hh>

#include <amdis/utility/UniqueBorderPartition.hpp>
#include <amdis/Initfile.hpp>
#include <amdis/Output.hpp>

namespace AMDiS {

#if HAVE_MPI

template <class Basis>
std::unique_ptr< ISTLCommunication<Basis> > ISTLCommunication<Basis>
::create(Basis const& basis, std::string const& prefix)
{
  std::string category = "default";
  Parameters::get(prefix + "->category", category);
  SolverType cat;
  auto const& gv = basis.gridView();
  int nProcs = gv.comm().size();

  if (category == "default")
  {
    if (nProcs == 1)
    {
      cat = SolverType::sequential;
    }
    else
    {
      // Use overlapping solver if grid has overlap
      if (gv.overlapSize(0) > 0)
        cat = SolverType::overlapping;
      else
      {
        // TODO(FM): Remove once nonoverlapping solvers are supported
        warning("Nonoverlapping solvers are currently not supported. "
          "Overlapping solver category chosen for grid with no overlap\n");
        cat = SolverType::overlapping;
        // cat = SolverType::nonoverlapping;
      }
    }
  }
  else if (category != "sequential" && nProcs == 1)
  {
    warning("Only one process detected. Solver category set to sequential\n");
    cat = SolverType::sequential;
  }
  else if (category == "sequential")
  {
    test_exit(nProcs == 1, "Solver category sequential is not supported in parallel\n");
    cat = SolverType::sequential;
  }
  else if (category == "overlapping")
  {
    if (gv.overlapSize(0) == 0)
      warning("Overlapping solver category chosen for grid with no overlap\n");
    cat = SolverType::overlapping;
  }
  else if (category == "nonoverlapping")
  {
    error_exit("Nonoverlapping solvers are currently not supported.");
    cat = SolverType::nonoverlapping;
  }
  else
  {
    error_exit("Unknown solver category\n");
  }

  return std::make_unique<Self>(basis, Dune::MPIHelper::getCommunicator(), cat);
}

template <class Basis>
void ISTLCommunication<Basis>
::buildParallelIndexSet(Basis const& basis, typename Super::PIS& pis)
{
  using Attribute = typename Dune::OwnerOverlapCopyAttributeSet::AttributeSet;
  using PType = Dune::PartitionType;
  using LI = typename Super::PIS::LocalIndex;

  auto const& gv = basis.gridView();

  // make disjoint partition of border entities
  using DataHandle = UniqueBorderPartition<Grid>;
  DataHandle borderEntities(gv.comm().rank(), gv.grid());
  for (int i = 0; i < borderEntities.numIterations(); ++i) {
    gv.communicate(borderEntities,
      Dune::InterfaceType::InteriorBorder_All_Interface,
      Dune::CommunicationDirection::ForwardCommunication);
  }

  if (gv.overlapSize(0) + gv.ghostSize(0) == 0)
    // TODO(FM): Add support for this special case
    error_exit("Using grids with ghostSize(0) + overlapSize(0) == 0 is not supported\n");

  auto lv = basis.localView();
  std::vector<bool> visited(basis.dimension(), false);
  GlobalBasisIdSet<Basis> dofIdSet(basis);

  pis.beginResize();
  for (auto const& e : Dune::elements(gv))
  {
    lv.bind(e);
    dofIdSet.bind(e);
    for (std::size_t i = 0; i < dofIdSet.size(); ++i)
    {
      LocalCommIdType localIndex = lv.index(i);
      if (!visited[localIndex]) {
        GlobalCommIdType globalId = dofIdSet.id(i);
        PType attribute = dofIdSet.partitionType(i);
        switch (attribute)
        {
        case PType::InteriorEntity:
          pis.add(globalId, LI(localIndex, Attribute::owner, true));
          break;
        case PType::BorderEntity:
          // mark border entity as owned iff it is part of the process's borderEntities set
          if (borderEntities.contains(dofIdSet.entityId(i)))
            pis.add(globalId, LI(localIndex, Attribute::owner, true));
          else
            pis.add(globalId, LI(localIndex, Attribute::overlap, true));
          break;
        case PType::OverlapEntity:
        case PType::FrontEntity:
          pis.add(globalId, LI(localIndex, Attribute::overlap, true));
          break;
        case PType::GhostEntity:
          pis.add(globalId, LI(localIndex, Attribute::copy, true));
          break;
        default:
          error_exit("Unknown partition type.");
        }

        visited[localIndex] = true;
      }
    }
    dofIdSet.unbind();
    lv.unbind();
  }
  pis.endResize();
}

#endif

} // end namespace AMDiS
